/* play-preferences-dialog.c
 *
 * Copyright 2022 Jonathan Blandford <jrb@gnome.org>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * SPDX-License-Identifier: GPL-3.0-or-later
 */


#include <glib/gi18n.h>

#include "contrib/gnome-languages.h"
#include "crosswords-app.h"
#include "crosswords-quirks.h"
#include "play-preferences-dialog.h"
#include "puzzle-set.h"
#include "puzzle-set-list.h"


enum
{
  DELETE_NEVER,
  DELETE_DAY,
  DELETE_WEEK,
  DELETE_MONTH,
};

struct _PlayPreferencesDialog
{
  AdwPreferencesWindow parent_instance;

  GSettings *settings;

  /* Template widgets */
  GtkWidget *guess_behavior_row;
  GtkWidget *switch_on_move_switch;
  GtkWidget *hide_solved_puzzles_switch;
  GtkWidget *delete_solved_puzzles_row;
  GtkWidget *show_all_languages_switch;
  GtkWidget *shown_sources_page;
  GtkWidget *shown_sources_list_box;
  GtkWidget *auto_download_switch;
};


static void       play_preferences_dialog_init          (PlayPreferencesDialog      *self);
static void       play_preferences_dialog_class_init    (PlayPreferencesDialogClass *klass);
static void       play_preferences_dialog_dispose       (GObject                    *object);
static gboolean   guess_behavior_row_get_mapping        (GValue                     *value,
                                                         GVariant                   *variant,
                                                         gpointer                    user_data);
static GVariant * guess_behavior_row_set_mapping        (const GValue               *value,
                                                         const GVariantType         *expected_type,
                                                         gpointer                    user_data);
static gboolean   delete_solved_puzzles_row_get_mapping (GValue                     *value,
                                                         GVariant                   *variant,
                                                         gpointer                    user_data);
static GVariant * delete_solved_puzzles_row_set_mapping (const GValue               *value,
                                                         const GVariantType         *expected_type,
                                                         gpointer                    user_data);
static void       shown_sources_page_activated_cb       (PlayPreferencesDialog      *self);
static void       shown_sources_back_cb                 (PlayPreferencesDialog      *self);
static GtkWidget* create_shown_sources_row              (PuzzleSet                  *puzzle_set,
                                                         PlayPreferencesDialog      *self);


G_DEFINE_TYPE (PlayPreferencesDialog, play_preferences_dialog, ADW_TYPE_PREFERENCES_WINDOW);


static void
play_preferences_dialog_init (PlayPreferencesDialog *self)
{
  GListModel *puzzle_set_list;
  GtkFilter *shown_sources_filter;
  g_autoptr (GtkFilterListModel) shown_sources_filter_model = NULL;

  gtk_widget_init_template (GTK_WIDGET (self));

#if DEVELOPMENT_BUILD
  gtk_widget_add_css_class (GTK_WIDGET (self), "devel");
#endif

  self->settings = g_settings_new ("org.gnome.Crosswords");

  /* Game Behavior */
  g_settings_bind_with_mapping (self->settings, "guess-advance-type",
                                self->guess_behavior_row, "selected",
                                G_SETTINGS_BIND_DEFAULT,
                                guess_behavior_row_get_mapping,
                                guess_behavior_row_set_mapping,
                                NULL, NULL);
  g_settings_bind (self->settings, "switch-on-move",
                   self->switch_on_move_switch, "active",
                   G_SETTINGS_BIND_DEFAULT);

  /* Solved Puzzles */
  g_settings_bind (self->settings, "hide-solved-puzzles",
                   self->hide_solved_puzzles_switch, "active",
                   G_SETTINGS_BIND_DEFAULT);
  g_settings_bind_with_mapping (self->settings, "delete-downloaded-puzzles",
                                self->delete_solved_puzzles_row, "selected",
                                G_SETTINGS_BIND_DEFAULT,
                                delete_solved_puzzles_row_get_mapping,
                                delete_solved_puzzles_row_set_mapping,
                                NULL, NULL);

  /* Puzzle Sources */
  g_settings_bind (self->settings, "show-all-languages",
                   self->show_all_languages_switch, "active",
                   G_SETTINGS_BIND_DEFAULT);

  puzzle_set_list = puzzle_set_list_get ();
  shown_sources_filter = puzzle_set_filter_new (PUZZLE_SET_FILTER_LANGUAGE,
                                                self->settings);
  /* Strangely, the filter_model takes ownership of the model and
   * filter, so we have to ref puzzle_set_list or it gets reaped.*/
  shown_sources_filter_model = gtk_filter_list_model_new (g_object_ref (puzzle_set_list),
                                                          shown_sources_filter);

  gtk_list_box_bind_model (GTK_LIST_BOX (self->shown_sources_list_box),
                           G_LIST_MODEL (shown_sources_filter_model),
                           (GtkListBoxCreateWidgetFunc) create_shown_sources_row,
                           self, NULL);

  g_settings_bind (self->settings, "auto-download-puzzle-sets",
                   self->auto_download_switch, "active",
                   G_SETTINGS_BIND_DEFAULT);
}

static void
play_preferences_dialog_class_init (PlayPreferencesDialogClass *klass)
{
  GObjectClass *object_class = G_OBJECT_CLASS (klass);
  GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (klass);

  object_class->dispose = play_preferences_dialog_dispose;

  gtk_widget_class_set_template_from_resource (widget_class, "/org/gnome/Crosswords/play-preferences-dialog.ui");

  gtk_widget_class_bind_template_child (widget_class, PlayPreferencesDialog, guess_behavior_row);
  gtk_widget_class_bind_template_child (widget_class, PlayPreferencesDialog, switch_on_move_switch);
  gtk_widget_class_bind_template_child (widget_class, PlayPreferencesDialog, hide_solved_puzzles_switch);
  gtk_widget_class_bind_template_child (widget_class, PlayPreferencesDialog, delete_solved_puzzles_row);
  gtk_widget_class_bind_template_child (widget_class, PlayPreferencesDialog, show_all_languages_switch);
  gtk_widget_class_bind_template_child (widget_class, PlayPreferencesDialog, shown_sources_page);
  gtk_widget_class_bind_template_child (widget_class, PlayPreferencesDialog, shown_sources_list_box);
  gtk_widget_class_bind_template_child (widget_class, PlayPreferencesDialog, auto_download_switch);

  gtk_widget_class_bind_template_callback (widget_class, shown_sources_page_activated_cb);
  gtk_widget_class_bind_template_callback (widget_class, shown_sources_back_cb);
}

static void
play_preferences_dialog_dispose (GObject *object)
{
  PlayPreferencesDialog *self;

  self = PLAY_PREFERENCES_DIALOG (object);

  g_clear_object (&self->settings);

  G_OBJECT_CLASS (play_preferences_dialog_parent_class)->dispose (object);
}

static gboolean
guess_behavior_row_get_mapping (GValue   *value,
                                GVariant *variant,
                                gpointer  user_data)
{
  const gchar *advance;

  advance = g_variant_get_string (variant, NULL);

  if (! g_strcmp0 (advance, "adjacent"))
    g_value_set_uint (value, (guint) QUIRKS_GUESS_ADVANCE_ADJACENT);
  else if (! g_strcmp0 (advance, "open"))
    g_value_set_uint (value, (guint) QUIRKS_GUESS_ADVANCE_OPEN);
  else if (! g_strcmp0 (advance, "open-in-clue"))
    g_value_set_uint (value, (guint) QUIRKS_GUESS_ADVANCE_OPEN_IN_CLUE);
  else
    {
      g_warning ("Unknown advance type: %s", advance);
      return FALSE;
    }

  return TRUE;
}

static GVariant *
guess_behavior_row_set_mapping (const GValue       *value,
                                const GVariantType *expected_type,
                                gpointer            user_data)
{
  GVariant *retval = NULL;

  /* FIXME(enum): Maybe we should use the enum nick for this */
  switch (g_value_get_uint (value))
    {
    case QUIRKS_GUESS_ADVANCE_ADJACENT:
      retval = g_variant_new_string ("adjacent");
      break;
    case QUIRKS_GUESS_ADVANCE_OPEN:
      retval = g_variant_new_string ("open");
      break;
    case QUIRKS_GUESS_ADVANCE_OPEN_IN_CLUE:
      retval = g_variant_new_string ("open-in-clue");
      break;
    default:
      g_assert_not_reached ();
    }

  return retval;
}


static gboolean
delete_solved_puzzles_row_get_mapping (GValue   *value,
                                       GVariant *variant,
                                       gpointer  user_data)
{
  const gchar *delete_solved_puzzles;

  delete_solved_puzzles = g_variant_get_string (variant, NULL);

  if (! g_strcmp0 (delete_solved_puzzles, "never"))
    g_value_set_uint (value, (guint) DELETE_NEVER);
  else if (! g_strcmp0 (delete_solved_puzzles, "day"))
    g_value_set_uint (value, (guint) DELETE_DAY);
  else if (! g_strcmp0 (delete_solved_puzzles, "week"))
    g_value_set_uint (value, (guint) DELETE_WEEK);
  else if (! g_strcmp0 (delete_solved_puzzles, "month"))
    g_value_set_uint (value, (guint) DELETE_MONTH);
  else
    {
      g_warning ("Unknown delete_solved_puzzles type: %s", delete_solved_puzzles);
      return FALSE;
    }

  return TRUE;
}

static GVariant *
delete_solved_puzzles_row_set_mapping (const GValue       *value,
                                       const GVariantType *expected_type,
                                       gpointer            user_data)
{
  GVariant *retval = NULL;

  /* FIXME(enum): Maybe we should use the enum nick for this */
  switch (g_value_get_uint (value))
    {
    case DELETE_NEVER:
      retval = g_variant_new_string ("never");
      break;
    case DELETE_DAY:
      retval = g_variant_new_string ("day");
      break;
    case DELETE_WEEK:
      retval = g_variant_new_string ("week");
      break;
    case DELETE_MONTH:
      retval = g_variant_new_string ("month");
      break;
    default:
      g_assert_not_reached ();
    }

  return retval;
}

static void
shown_sources_page_activated_cb (PlayPreferencesDialog *self)
{
  adw_preferences_window_present_subpage (ADW_PREFERENCES_WINDOW (self), self->shown_sources_page);
}

static void
shown_sources_back_cb (PlayPreferencesDialog *self)
{
  adw_preferences_window_close_subpage (ADW_PREFERENCES_WINDOW (self));
}

#define SHOWN_SWITCH_DATA "shown-switch-data"

static void
shown_switch_toggled (GtkWidget             *shown_switch,
                      GParamSpec            *pspec,
                      PlayPreferencesDialog *self)
{
  GVariantBuilder builder;
  g_autoptr (GVariant) old_show_list = NULL;
  gboolean shown;
  const gchar *id;
  gboolean found_id = FALSE;

  old_show_list = g_settings_get_value (self->settings, "shown-puzzle-sets");
  shown = gtk_switch_get_state (GTK_SWITCH (shown_switch));
  id = g_object_get_data (G_OBJECT (shown_switch), SHOWN_SWITCH_DATA);

  g_variant_builder_init (&builder, G_VARIANT_TYPE_STRING_ARRAY);

  for (guint i = 0; i < g_variant_n_children (old_show_list); i++)
    {
      g_autoptr (GVariant) child = NULL;

      child = g_variant_get_child_value (old_show_list, i);
      if (g_strcmp0 (id, g_variant_get_string (child, NULL)) == 0)
        {
          if (!shown)
            found_id = TRUE;
          continue;
        }
      g_variant_builder_add (&builder, "s", g_variant_get_string (child, NULL));
    }
  /* We use found_id to guard against accidentally adding this
   * twice. */
  if (shown && !found_id)
    g_variant_builder_add (&builder, "s", id);

  g_settings_set_value (self->settings, "shown-puzzle-sets",
                        g_variant_builder_end (&builder));
}

static void
add_tag (GString       *subtitle,
         ConfigSetTags  tags,
         ConfigSetTags  tag,
         const gchar   *tag_str,
         gboolean      *show_slash)
{
  if (tags & tag)
    {
      if (*show_slash)
        g_string_append (subtitle, " / ");

      g_string_append (subtitle, tag_str);
      *show_slash = TRUE;
    }
}

static gchar *
create_subtitle (PuzzleSet *puzzle_set)
{
  const char *locale;
  g_autofree gchar *language = NULL;
  GString *subtitle = g_string_new (NULL);
  ConfigSetTags tags;
  gboolean show_slash = FALSE;

  locale = puzzle_set_get_locale (puzzle_set);
  language = gnome_get_country_from_locale (locale, NULL);
  tags = puzzle_set_get_tags (puzzle_set);

  if (language)
    g_string_append (subtitle, language);

  /* Add the bullet if we have something we're going to show */
  if (tags & (CONFIG_SET_TAGS_MINI |
              CONFIG_SET_TAGS_JUMBO))
    g_string_append (subtitle, " • ");

  add_tag (subtitle, tags, CONFIG_SET_TAGS_MINI, _("Mini"), &show_slash);
  if (tags & CONFIG_SET_TAGS_REGULAR)
    {
      if (tags & (CONFIG_SET_TAGS_MINI | CONFIG_SET_TAGS_JUMBO))
        add_tag (subtitle, tags, CONFIG_SET_TAGS_REGULAR, _("Regular"), &show_slash);
    }
  add_tag (subtitle, tags, CONFIG_SET_TAGS_JUMBO, _("Jumbo"), &show_slash);

  /* Add the bullet if we have something we're going to show */
  if (tags & (CONFIG_SET_TAGS_ACROSTIC |
              CONFIG_SET_TAGS_ARROWWORD |
              CONFIG_SET_TAGS_BARRED |
              CONFIG_SET_TAGS_CROSSWORD |
              CONFIG_SET_TAGS_CRYPTIC |
              CONFIG_SET_TAGS_FILIPPINE))
    g_string_append (subtitle, " • ");

  show_slash = FALSE;
  add_tag (subtitle, tags, CONFIG_SET_TAGS_ACROSTIC, _("Acrostic Puzzles"), &show_slash);
  add_tag (subtitle, tags, CONFIG_SET_TAGS_ARROWWORD, _("Arrowword Puzzles"), &show_slash);
  add_tag (subtitle, tags, CONFIG_SET_TAGS_BARRED, _("Barred Puzzles"), &show_slash);
  add_tag (subtitle, tags, CONFIG_SET_TAGS_CROSSWORD, _("Crosswords"), &show_slash);
  add_tag (subtitle, tags, CONFIG_SET_TAGS_CRYPTIC, _("Cryptic Crosswords"), &show_slash);
  add_tag (subtitle, tags, CONFIG_SET_TAGS_FILIPPINE, _("Filippines"), &show_slash);

  if (puzzle_set_get_auto_download (puzzle_set))
    {
      g_string_append (subtitle, " • ");
      g_string_append (subtitle, _("Automatic Downloader"));
    }

  return g_string_free (subtitle, FALSE);
}

static GtkWidget *
create_shown_sources_row (PuzzleSet             *puzzle_set,
                          PlayPreferencesDialog *self)
{
  GtkWidget *shown_switch;
  GtkWidget *row;
  g_autofree gchar *subtitle = NULL;

  subtitle = create_subtitle (puzzle_set);

  shown_switch = gtk_switch_new ();
  gtk_switch_set_state (GTK_SWITCH (shown_switch), puzzle_set_get_shown (puzzle_set));
  gtk_widget_set_can_focus (shown_switch, FALSE);
  gtk_widget_set_valign (shown_switch, GTK_ALIGN_CENTER);
  row = g_object_new (ADW_TYPE_ACTION_ROW,
                      "title", puzzle_set_get_short_name (puzzle_set),
                      "subtitle", subtitle,
                      "activatable", true,
                      "activatable-widget", shown_switch,
                      NULL);
  adw_action_row_add_suffix (ADW_ACTION_ROW (row),
                             shown_switch);
  g_object_set_data (G_OBJECT (shown_switch), SHOWN_SWITCH_DATA, (gpointer) puzzle_set_get_id (puzzle_set));
  g_signal_connect (shown_switch, "notify::state", G_CALLBACK (shown_switch_toggled), self);

  return row;
}

/* Public functions */

GtkWidget *
play_preferences_dialog_new (GtkWindow *parent_window,
                             gboolean   display_shown_sources_page)
{
  GtkWidget *dialog;

  g_return_val_if_fail (GTK_IS_WINDOW (parent_window), NULL);

  dialog = g_object_new (PLAY_TYPE_PREFERENCES_DIALOG,
                         "application", CROSSWORDS_APP_DEFAULT,
                         "transient-for", parent_window,
                         NULL);
  if (display_shown_sources_page)
    shown_sources_page_activated_cb (PLAY_PREFERENCES_DIALOG (dialog));

  return dialog;
}
